from collections import namedtuple
import operator
from functools import reduce
from itertools import chain
import re
from funcparserlib.lexer import TokenSpec, make_tokenizer
from funcparserlib.parser import Parser, some, tok, forward_decl, maybe, many, finished

UnaryExpr = namedtuple('UnaryExpr', 'operator value')
BinaryExpr = namedtuple('BinaryExpr', 'operator left right')
VariableReference = namedtuple('VariableReference', 'name')
ArrayVariableReference = namedtuple('ArrayVariableReference', 'name index')
FunctionReference = namedtuple('FunctionReference', 'name')
FunctionCall = namedtuple('FunctionCall', 'name parameters')

TypeDeclaration = namedtuple('TypeDeclaration', 'name parent')
NativeFunction = namedtuple('NativeFunction', 'constant name parameters returns')
Variable = namedtuple('Variable', 'name type array constant value')
class Variables(list): pass

class _UnaryOperator(str): pass

arithmetic_ops = [
	_UnaryOperator('-'),
	'-',
	'+',
	'/',
	'*',
]

comparison_ops = [
	'>=',
	'<=',
	'>',
	'<',
]

logical_ops = [
	_UnaryOperator('not'),
	'and',
	'or'
	'==',
	'!=',
]

_computation_operators = [
	*arithmetic_ops,
	*comparison_ops,
	*logical_ops,
]

_operators = [
	*_computation_operators,
	'=',
]

_brackets = [
	'(',
	')',
	'[',
	']',
]

op_regex = r'(%s)(?=[^a-zA-Z_]|$)' % '|'.join(f'({re.escape(x)})' for x in _operators)

_specification = (
	TokenSpec('whitespace', r'\s+'),
	TokenSpec('comment', r'//.*\n'),
	TokenSpec('integer', r'[0-9]+'),
	TokenSpec('decimal', r'([0-9]+\.([0-9]+)?)|(([0-9]+)\.[0-9]+)'),
	TokenSpec('hex', r'\$[0-9a-fA-F]+'),
	TokenSpec('boolean', r'true|false'),
	TokenSpec('null', r'null'),
	TokenSpec('string', r'"((\\")|[^\n])*"'),
	TokenSpec('op', op_regex),
	TokenSpec('bracket', '|'.join(f'({re.escape(x)})' for x in _brackets)),
	TokenSpec('comma', ','),
	TokenSpec('identifier', r'[_a-zA-Z][_a-zA-Z0-9]*'),
)

_useless = {'whitespace', 'comment'}
_tokenizer = make_tokenizer(_specification)

def tokenize(text):
	return (x for x in _tokenizer(text) if x.type not in _useless)

def unpack_parameters(f):
	return lambda x: f(*x)

class _Repeat(Parser):

	def __init__(self, parser):
		super(_Repeat, self).__init__(self._run)
		self.parser = parser

	def until(self, parser):
		self.condition = parser
		return self

	def _run(self, tokens, state):
		result = []
		temp_state = state

		while True:
			value, temp_state = self.parser.run(tokens, temp_state)
			result.append(value)

			try:
				# final_value, temp_state = self.condition.run(tokens, temp_state)
				self.condition.run(tokens, temp_state)
			except:
				pass
			else:
				break

		# if isinstance(final_value, _Ignored):
		# 	return result, temp_state
		# else:
		# 	return result, (temp_state + final_value)

		return result, temp_state


def repeat(parser):
	return _Repeat(parser)

any_token = some(lambda _: True)

integer = tok('integer') >> int
decimal = tok('decimal') >> float
hexadecimal = tok('hex') >> (lambda x: int(x, 16))
number = integer | decimal | hexadecimal
boolean = tok('boolean') >> bool
null = tok('null') >> (lambda _: None)
string = tok('string') >> (lambda s: s[1:-1])
identifier = tok('identifier')
comma = tok('comma')

def keyword(name):
	return tok('identifier', name)

def op(name):
	return tok('op', name)

def bracket(name):
	return tok('bracket', name)

def parenthesised(parser):
	return -bracket('(') + parser + -bracket(')')

def bracketed(parser):
	return -bracket('[') + parser + -bracket(']')

def separated_list(separator, parser):
	def make_list(x):
		result = [x[0], *x[1]] if x else []
		return result

	return maybe(parser + many(separator + parser)) >> make_list

def make_unary(x):
	operators, value = x

	for operator in reversed(operators):
		value = UnaryExpr(operator, value)

	return value

def make_binary(x):
	iterator = iter(x)
	result = next(iterator)

	for op, expr in iterator:
		result = BinaryExpr(op, result, expr)

	return result

def make_parameter(type, name):
	return Variable(name, type, False, False, None)

def make_variable(constant, type, array, name, value):
	return Variable(name, type, array, constant, value)

def make_variables(variables):
	x = Variables(variables)
	print('make_variables returns', x.__class__, len(x))
	return x

expression = forward_decl()

variable = identifier >> VariableReference
array = (identifier + bracketed(expression)) >> unpack_parameters(ArrayVariableReference)
function = (-keyword('function') + identifier) >> FunctionReference

call = (identifier + parenthesised(separated_list(-comma, expression))) >> unpack_parameters(FunctionCall)

primary = number | boolean | null | string | array | function | call | variable | parenthesised(expression)

def make_(start, operators):
	current = start

	for operator in operators:
		if isinstance(operator, _UnaryOperator):
			current = (many(op(operator)) + current) >> make_unary
		else:
			current = separated_list(op(operator), current) >> make_binary

	return current

expression.define(make_(primary, _computation_operators))

variable_modifiers = maybe(keyword('constant')) >> bool
variable_array_flag = maybe(keyword('array')) >> bool
variable_initializer = maybe(-op('=') + expression)
variable_declaration = (variable_modifiers + identifier + variable_array_flag + identifier + variable_initializer) >> unpack_parameters(make_variable)

local_variable_declaration = -keyword('local') + variable_declaration

# globals_block = \
# 	-keyword('globals') + \
# 		many(variable_declaration) + \
# 	-keyword('endglobals')

globals_block = (
	-keyword('globals') +
		repeat(variable_declaration).until(keyword('endglobals')) +
	-keyword('endglobals')
) >> make_variables


type_declaration = (-keyword('type') + identifier + -keyword('extends') + identifier) >> unpack_parameters(TypeDeclaration)

no_parameters = keyword('nothing') >> (lambda _: [])
parameter_declaration = (identifier + identifier) >> unpack_parameters(make_parameter)
parameters_list = separated_list(-comma, parameter_declaration)
parameters_declaration = no_parameters | parameters_list

native_declaration = \
	(maybe(keyword('constant')) >> bool) + \
	-keyword('native') + \
	identifier + \
	-keyword('takes') + \
	parameters_declaration + \
	-keyword('returns') + \
	identifier \
	>> unpack_parameters(NativeFunction)

jass_function = \
	-keyword('function') + \
		identifier + \
		-repeat(any_token).until(keyword('endfunction')) + \
	-keyword('endfunction')

jass = many(type_declaration | globals_block | native_declaration | jass_function) + -finished
