from io import StringIO
from functools import singledispatch
from itertools import chain
from jass_parser import TypeDeclaration, Variables, Variable, NativeFunction, UnaryExpr, BinaryExpr, FunctionCall, jass, tokenize

def to_pascal_variable_type(s):
	d = {
		'integer': 'Integer',
		'real': 'Single',
		'boolean': 'Boolean',
		'string': 'String',
		'code': 'TJassCallback',
		'handle': 'TJassObjectHandle',
	}

	return d.get(s) or ('H%s' % s.upper())

def to_pascal_parameter_type(s):
	return to_pascal_variable_type(s)

def to_pascal_result_type(s):
	return to_pascal_variable_type(s)

def to_pascal_raw_parameter_type(s):
	d = {
		'integer': 'Int32',
		'real': 'PSingle',
		'boolean': 'LongBool',
		'string': 'PJassString',
	}

	return d.get(s) or to_pascal_variable_type(s)

def to_pascal_raw_result_type(s):
	d = {
		'integer': 'Int32',
		'real': 'TSingle',
		'boolean': 'LongBool',
		'string': 'TJassStringHandle',
	}

	return d.get(s) or to_pascal_variable_type(s)

_pascal_keywords = {
	'file', 'of', 'array',
	'label', 'goto',
	'true', 'false',
	'and', 'not', 'or',
	'if', 'then', 'else'
	'begin', 'end',
	'function', 'procedure',
	'var', 'const', 'constref', 'out',
	'type', 'class', 'record', 'object',
	'uses', 'interface', 'implementation',
}

def to_pascal_identifier(s):
	if s.lower() in _pascal_keywords:
		return f'{s}_'
	else:
		return s

@singledispatch
def converter(x, stream):
	raise Exception(f'Value cannot be converted: "{x}".')

@converter.register(int)
@converter.register(float)
@converter.register(type(None))
def convert_number(x, context):
	context.output.write(str(x))

@converter.register(bool)
def convert_bool(x, context):
	context.output.write('True' if x else 'False')

@converter.register(list)
def convert_list(items, context):
	for x in items:
		converter(x, context)

@converter.register(TypeDeclaration)
def convert_type(type, context):
	if context.stage == STAGE_INTERFACE:
		context.output.write('\t{0} = {1};\n'.format(
			to_pascal_variable_type(type.name),
			to_pascal_variable_type(type.parent)
		))

@converter.register(Variable)
def convert_variable(variable, context):
	if context.stage == STAGE_INTERFACE:
		context.output.write('\t{0}: {1};\n'.format(
			to_pascal_identifier(variable.name),
			to_pascal_variable_type(variable.type)
		))
	elif context.stage == STAGE_JASS_INITIALIZATION:
		if variable.value != None:
			context.output.write('\t%s:= ' % to_pascal_identifier(variable.name))
			converter(variable.value, context)
			context.output.write(';\n')

def generate_native_function_type(native, context):
	parameters = '; '.join(
		f'{to_pascal_identifier(parameter.name)}: {to_pascal_raw_parameter_type(parameter.type)}'
			for parameter in native.parameters
	)
	if parameters: parameters = f' ({parameters})'
	returns = f': {to_pascal_raw_result_type(native.returns)}' if native.returns != 'nothing' else ''
	keyword = 'procedure' if native.returns == 'nothing' else 'function'

	context.output.write(f'type T{native.name}FunctionPrototype = {keyword}{parameters}{returns}; cdecl;\n')

def generate_native_function_variable(native, context):
	context.output.write(f'var _{native.name}: T{native.name}FunctionPrototype;\n')

def generate_native_function_wrapper_header(native, context):
	keyword = 'procedure' if native.returns == 'nothing' else 'function'
	context.output.write(f'{keyword} {to_pascal_identifier(native.name)}')

	if len(native.parameters) > 0:
		context.output.write('(')

		context.output.write('; '.join(
			f'{to_pascal_identifier(parameter.name)}: {to_pascal_parameter_type(parameter.type)}'
				for parameter in native.parameters
		))

		context.output.write(')')

	if native.returns != 'nothing':
		context.output.write(f': {to_pascal_result_type(native.returns)}')

	context.output.write(';\n')

def generate_native_function_wrapper_prologue(native, context):
	parameters = ''.join(
		f'\tj{parameter.name}: PJassString;\n'
			for parameter in native.parameters
				if parameter.type == 'string'
	)

	if native.returns == 'string':
		extra_result_var = '\thResult: TJassStringHandle;\n'
	else:
		extra_result_var = ''

	if parameters or extra_result_var:
		context.output.write('var\n')
		context.output.write(parameters)
		context.output.write(extra_result_var)

	context.output.write('begin\n')

	parameters_initialization = ''.join(
		f'\tj{parameter.name}:= CreateJassString(PChar({to_pascal_identifier(parameter.name)}));\n'
			for parameter in native.parameters
				if parameter.type == 'string'
	)
	context.output.write(parameters_initialization)
	

def make_native_parameter_conversion(parameter):
	if parameter.type == 'string':
		return f'j{parameter.name}'
	elif parameter.type == 'real':
		return f'@{to_pascal_identifier(parameter.name)}'
	else:
		return to_pascal_identifier(parameter.name)

def make_native_result_conversion(value_type, value):
	if value_type == 'real':
		return f'{value}.Value'
	else:
		return value

def make_native_call(native):
	return f'''_{native.name}({', '.join(
		make_native_parameter_conversion(parameter)
			for parameter in native.parameters
	)})'''

def generate_native_function_wrapper_body(native, context):
	context.output.write('\t')

	if native.returns == 'string':
		context.output.write('hResult:= ')
	elif native.returns != 'nothing':
		context.output.write('Result:= ')

	context.output.write(
		make_native_result_conversion(
			native.returns,
			make_native_call(native)
		)
	)

	context.output.write(';\n')

	if native.returns == 'string':
		context.output.write('\tResult:= GetStringFromHandle(hResult);\n')

def generate_native_function_wrapper_epilogue(native, context):
	if native.returns == 'string':
		context.output.write('\tDecrementStringReferences(hResult);\n')

	finalizing_lines = ''.join(
		f'\tDestroyJassString(j{parameter.name});\n'
			for parameter in native.parameters
				if parameter.type == 'string'
	)
	context.output.write(finalizing_lines)
	context.output.write('end;\n')

def generate_native_function_wrapper(native, context):
	generate_native_function_wrapper_header(native, context)
	generate_native_function_wrapper_prologue(native, context)
	generate_native_function_wrapper_body(native, context)
	generate_native_function_wrapper_epilogue(native, context)

@converter.register(NativeFunction)
def convert_native(native, context):
	if context.stage == STAGE_INTERFACE:
		generate_native_function_wrapper_header(native, context)
	elif context.stage == STAGE_IMPLEMENTATION:
		generate_native_function_type(native, context)
		generate_native_function_variable(native, context)
		context.output.write('\n')
		generate_native_function_wrapper(native, context)
	elif context.stage == STAGE_JASS_INITIALIZATION:
		context.output.write(f'''\tPointer(_{native.name}):= GetJassNative('{native.name}');\n''')

@converter.register(FunctionCall)
def convert_call(function, context):
	context.output.write(f'{to_pascal_identifier(function.name)}(')

	for index, param in enumerate(function.parameters):
		converter(param, context)

		if index < len(function.parameters) - 1:
			context.output.write(', ')

	context.output.write(')')

def convert_child_expression(child, parent, context):
	if isinstance(child, (UnaryExpr, BinaryExpr)):
		wrap_child = child.operator < parent.operator
	else:
		wrap_child = False

	if wrap_child: context.output.write('(')
	converter(child, context)
	if wrap_child: context.output.write(')')

@converter.register(BinaryExpr)
def convert_expression(expression, context):
	convert_child_expression(expression.left, expression, context)
	context.output.write(expression.operator)
	convert_child_expression(expression.right, expression, context)

@converter.register(UnaryExpr)
def convert_expression(expression, context):
	context.output.write(expression.operator)
	convert_child_expression(expression.value, expression, context)

def parse_header(script):
	parsed = jass.parse(list(tokenize(script)))

	types = [x for x in parsed if isinstance(x, TypeDeclaration)]
	variables = list(chain(*(x for x in parsed if isinstance(x, Variables))))
	natives = [x for x in parsed if isinstance(x, NativeFunction)]

	return types, variables, natives

if __name__ == '__main__':
	with open('./common.j', 'rt') as f:
		script = f.read()

	types, variables, natives = parse_header(script)

	stream = StringIO()

	stream.write('''unit jass_common_j;

interface

uses
	JassAPI;

''')

	class Context: pass

	STAGE_INTERFACE = 0
	STAGE_IMPLEMENTATION = 1
	STAGE_JASS_INITIALIZATION = 2

	context = Context()
	context.output = stream
	context.stage = STAGE_INTERFACE

	stream.write('type\n')
	stream.write('\tTJassCallback = function: Dword; stdcall;\n')

	if len(types) > 0:
		for x in types:
			converter(x, context)

	stream.write('\n')

	if len(variables) > 0:
		stream.write('var\n')
		for x in variables:
			converter(x, context)
		stream.write('\n')

	if len(natives) > 0:
		for x in natives:
			converter(x, context)
		stream.write('\n')

	stream.write('implementation\n\n')
	context.stage = STAGE_IMPLEMENTATION

	for x in types:
		converter(x, context)

	for x in variables:
		converter(x, context)

	for x in natives:
		converter(x, context)
		stream.write('\n')

	stream.write('procedure OnMainBeingCalled({%H-}AParameter: Pointer); stdcall;\n')
	stream.write('begin\n')
	context.stage = STAGE_JASS_INITIALIZATION

	for x in types:
		converter(x, context)

	for x in natives:
		converter(x, context)

	for x in variables:
		converter(x, context)

	stream.write('end;\n\n')

	stream.write('begin\n')
	stream.write('\tAddEventHandler(jeMainBeingCalled, @OnMainBeingCalled, nil);\n')
	stream.write('end.\n')

	with open('./common.pas', 'wt') as f:
		f.write(stream.getvalue())
